import sys
import argparse
import os
import json
import yaml
import xmltodict
import re
import traceback
import time
import random
import string
from collections import deque
from types import FunctionType

def _x(ds) :
    return json.dumps(ds,indent=2,sort_keys=True)

def _y(ds) :
    return yaml.dump(ds,default_flow_style=False)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--infile", dest="infile", help="input file")
    parser.add_argument("-t", "--srctype", dest="srctype", default="JSON", help="JSON,YAML or XML")
    parser.add_argument("-c", "--code", dest="code", default="_x(_)", help="code to compile. may be a file.")
    parser.add_argument("-i", "--indent", dest="indent", default=4, help="how many spaces for indent. default 4.")
    parser.add_argument("-m", "--module", dest="module", help="import modules.")
    parser.add_argument("-F", "--functionize", dest="func", action="store_true", default=False, help="wrap code into an internal function",)
    parser.add_argument("-e", "--exec", dest="exec", action="store_true", default=False, help="execute code block. all non value display purpose",)
    parser.add_argument("-I", "--interactive", dest="interactive", action="store_true", default=False, help="interactive mode",)
    parser.add_argument("-d", "--dotexpand", dest="dotexpand", action="store_true", default=False, help="expand . for keys",)
    parser.add_argument("-v", "--showcmd", dest="showcmd", action="store_true", default=False, help="show statement to run",)
    parser.add_argument("-X", "--debug", dest="debug", action="store_true", default=False, help="debug mode",)
    _x_args = parser.parse_args()

    def dotexpand(_x_sin) :
        chgsin = _x_sin
        _x_m = re.search(r"(\w+|\]|\))(\.\w+)+",chgsin)
        while _x_m :
            before = chgsin[:_x_m.start()]
            end = chgsin[_x_m.end():]
            chain = _x_m.group(0)
            newchain = ""
            for w in chain.split(".") :
                if newchain :
                    newchain += "['"+ w + "']"
                else :
                    newchain += w
            chgsin = before + newchain + end
            _x_m = re.search(r"(\w+|\]|\))(\.\w+)+",chgsin)
        return chgsin

    if _x_args.infile:
        if not os.path.isfile(_x_args.infile):
            print("# {} not exists.".format(_x_args.infile))
        with open(_x_args.infile, "r") as f:
            INPUT = f.read()
    else:
        INPUT = sys.stdin.read()
    INPUT = INPUT.strip()
    if not INPUT :
        INPUT = json.dumps({})
    if _x_args.debug :
        print("# INPUT = [{}]".format(INPUT))
    try :
       if _x_args.srctype.upper() == "JSON" :
            _ = json.loads(INPUT)
       elif _x_args.srctype.upper() == "YAML" :
            _ = yaml.safe_load(INPUT)
       elif _x_args.srctype.upper() == "XML" :
            _ = xmltodict.parse(INPUT)
       else :
        print("# unsupported file type.")
        return -1
    except :
        print("# invalid JSON/YAML/XML.")
        traceback.print_exc()
        return -1


    sys.stdin = sys.__stdin__

    _x_dotkey = _x_args.dotexpand
    if _x_args.interactive :
        _x_history = deque(maxlen=200)
        _x_cmd0=""
        print("$[dsq]: data loaded.")
        _x_oldsin=""
        _x_block=False
        while True :
            _x_doeval=True
            res=""
            err=""
            _x_sin = input("$[dsq]: ")
            if _x_sin.startswith("'''")  :
                if _x_block :
                    _x_sin = _x_oldsin 
                    _x_oldsin=""
                    _x_block=False
                    _x_doeval = False
                else :
                    _x_block = True
                    continue
            if _x_block :
                if _x_oldsin :
                    _x_oldsin += "\n" + _x_sin
                else :
                    _x_oldsin += _x_sin
                continue
            if not _x_sin :
                continue
            lastc =[int(ord(c)) for c in _x_sin][-1]
            if lastc and lastc in [65,66,67,68] : 
                _x_sin = "\\hist"
            if _x_sin in ["quit()","\\q"] :
                break
            if _x_sin in ["dotkey"] :
                _x_dotkey = True
                continue
            if _x_sin in ["nodotkey","no dotkey"] :
                _x_dotkey = False
                continue
            if _x_sin in ["\\hist","\\history"] :
                if not _x_history :
                    print("# no history found.")
                for i,cmd in enumerate(_x_history) :
                    print("# {:3} : {}".format(i,cmd))
                continue
            _x_m1 = re.match(r"\\r (\d+)",_x_sin) 
            _x_m2 = re.match(r"\\(\d+)",_x_sin) 
            if _x_m1 or _x_m2:
                if _x_m1 :
                    ix = int(_x_m1.group(1))
                if _x_m2 :
                    ix = int(_x_m2.group(1))
                if ix < len(_x_history) :
                    _x_sin = _x_history[ix]
                else :
                    continue
            _x_cmd0 = None
            if _x_dotkey and "." in _x_sin :
                _x_cmd0 = _x_sin
                _x_sin = dotexpand(_x_sin)
            try :
                if _x_cmd0 :
                    _x_history.append(_x_cmd0)
                elif not _x_sin.startswith("\\") :
                    _x_history.append(_x_sin)
                if _x_sin.startswith("\\") :
                    res = ""
                    err = "# command not recognized."
                else :
                    if re.match(r"\S+=\S*",_x_sin) :
                        _x_doeval=False
                    if _x_doeval :
                        if _x_args.showcmd :
                            print("# eval : {}".format(_x_sin))
                        res = eval(_x_sin)
                    else :
                        if _x_args.showcmd :
                            print("# exec : \n{}".format(_x_sin))
                        res = exec(_x_sin)
            except :
                err = traceback.format_exc()
            finally :
                if res :
                    print(res)
                if err :
                    print(err)
            time.sleep(0.1)
        return 0

    if os.path.isfile(_x_args.code) :
        code = open(_x_args.code,"r").read()
    else :
        code = _x_args.code

    if _x_args.func :
        fname = "".join([random.choice(string.ascii_letters) for _ in range(20)])
        newcode = ""
        if _x_args.module :
            for m in _x_args.module.split(",")  :
                if m :
                    m.strip()
                    if m.startswith("from ") :
                        newcode += m + "\n"
                    else :
                        newcode += "import " + m + "\n"
        newcode += "def {}(_) :\n".format(fname)
        for ln in code.splitlines() :
            newcode += " "*(int(_x_args.indent)) + ln.rstrip()
        code = newcode
        if _x_args.debug :
            print("# code to compile :")
            print(code)
        xcode = compile(code,"<string>","exec")
        cobj = None
        for c in xcode.co_consts :
            if c and type(c) not in [str,int,tuple] :
                cobj = c
                break
        if not cobj :
            print("# no code object found.")
            sys.exit(-1)
    
        xfunc = FunctionType(cobj, globals())
        print(xfunc(_))
        return(0)

    if _x_args.exec :
        if _x_args.dotexpand :
            code = dotexpand(code)
        if _x_args.showcmd :
            print("# exec : {}".format(code))
        exec(code)
    else :
        if _x_args.dotexpand :
            code = dotexpand(code)
        if _x_args.showcmd :
            print("# eval : {}".format(code))
        res = eval(code)
        print(res)


if __name__ == "__main__" :
    main()
